import os
from typing import Union


def is_exe(path: str) -> bool:
    return os.path.isfile(path) and os.access(path, os.X_OK)


def remove(
    folder: str,
    path: str,
    separator: str = os.pathsep,
    remove_duplicates: bool = True,
    remove_non_folders: bool = True,
    remove_non_abs: bool = True,
) -> str:
    path_elements = path.split(separator)
    if folder != "":
        path_elements = [x for x in path_elements if x != folder]
    new_path = separator.join(path_elements)
    new_path = clean(
        new_path,
        separator,
        remove_duplicates=remove_duplicates,
        remove_non_folders=remove_non_folders,
        remove_non_abs=remove_non_abs,
    )
    return new_path


def add(
    folder: str,
    path: str,
    head: bool,
    separator: str = os.pathsep,
    remove_duplicates: bool = True,
    remove_non_folders: bool = True,
    remove_non_abs: bool = True,
) -> str:
    path = remove(
        folder=folder,
        path=path,
        separator=separator,
        remove_duplicates=remove_duplicates,
        remove_non_folders=remove_non_folders,
        remove_non_abs=remove_non_abs,
    )
    path_elements = path.split(separator)
    if folder != "":
        if head:
            path_elements.insert(0, folder)
        else:
            path_elements.append(folder)
    path = separator.join(path_elements)
    path = clean(
        path,
        separator,
        remove_duplicates=remove_duplicates,
        remove_non_folders=remove_non_folders,
        remove_non_abs=remove_non_abs,
    )
    return path


def do_remove_non_abs(path: str, separator: str = os.pathsep) -> str:
    return separator.join([x for x in path.split(separator) if os.path.isabs(x)])


def do_remove_non_folders(path: str, separator: str = os.pathsep) -> str:
    return separator.join([x for x in path.split(separator) if os.path.isdir(x)])


def do_remove_duplicates(path: str, separator: str = os.pathsep) -> str:
    s = set()
    unique = []
    for path_element in path.split(separator):
        if path_element not in s:
            s.add(path_element)
            unique.append(path_element)
    return separator.join(unique)


def clean(
    path: str,
    separator: str = os.pathsep,
    remove_duplicates: bool = True,
    remove_non_folders: bool = True,
    remove_non_abs: bool = True,
) -> str:
    """
    returns a reduced version of the path. This means without repetition and without parts
    which are not folders.
    :param remove_non_abs: 
    :param remove_non_folders: 
    :param remove_duplicates: 
    :param path: 
    :param separator: 
    :return: 
    """
    if remove_duplicates:
        path = do_remove_duplicates(path, separator)
    if remove_non_folders:
        path = do_remove_non_folders(path, separator)
    if remove_non_abs:
        path = do_remove_non_abs(path, separator)
    return path


def find_in_path(path: str, app: str, separator: str = os.pathsep, strict: bool = False) -> Union[None, str]:
    """
    Return the full path if found, None otherwise
    :param path: 
    :param app: 
    :param separator: 
    :param strict: 
    :return: 
    """
    for path_element in path.split(separator):
        if strict:
            assert os.path.isabs(path_element), "Not absolute"
            assert os.path.isdir(path_element), "Not a folder"
        full_path_app = os.path.join(path_element, app)
        if is_exe(full_path_app):
            return full_path_app
    return None


def find_in_standard_path(app: str) -> Union[None, str]:
    """
    Find in the standard path
    :param app:
    :return:
    """
    return find_in_path(
        path=os.environ["PATH"],
        app=app,
    )



"""
The default group of operations that pypathutil has
"""

from pytconf.config import register_endpoint, register_function_group

import pypathutil.version
from pypathutil import common
from pypathutil.configs import ConfigFolder, ConfigPath, ConfigHead, ConfigSeparator, ConfigOps

GROUP_NAME_DEFAULT = "default"
GROUP_DESCRIPTION_DEFAULT = "all pypathutil commands"


def register_group_default():
    """
    register the name and description of this group
    """
    register_function_group(
        function_group_name=GROUP_NAME_DEFAULT,
        function_group_description=GROUP_DESCRIPTION_DEFAULT,
    )


@register_endpoint(
    group=GROUP_NAME_DEFAULT,
)
def version() -> None:
    """
    Print version
    """
    print(pypathutil.version.VERSION_STR)


@register_endpoint(
    configs=[
        ConfigFolder,
        ConfigPath,
        ConfigHead,
        ConfigSeparator,
        ConfigOps,
    ]
)
def add() -> None:
    """
    add two paths together
    """
    new_path = common.add(
        path=ConfigPath.path,
        folder=ConfigFolder.folder,
        head=ConfigHead.head,
        separator=ConfigSeparator.separator,
        remove_duplicates=ConfigOps.remove_duplicates,
        remove_non_folders=ConfigOps.remove_non_folders,
        remove_non_abs=ConfigOps.remove_non_abs,
    )
    print(new_path)


@register_endpoint(
    configs=[
        ConfigSeparator,
        ConfigOps,
        ConfigPath,
    ]
)
def clean() -> None:
    """
    clean a path, removing elements which repeat or are not valid paths
    """
    path = common.clean(
        path=ConfigPath.path,
        separator=ConfigSeparator.separator,
        remove_duplicates=ConfigOps.remove_duplicates,
        remove_non_folders=ConfigOps.remove_non_folders,
        remove_non_abs=ConfigOps.remove_non_abs,
    )
    print(path)


@register_endpoint(
    configs=[
        ConfigOps,
        ConfigSeparator,
        ConfigPath,
        ConfigFolder,
    ]
)
def remove() -> None:
    """ remove element from a path """
    new_path = common.remove(
        path=ConfigPath.path,
        folder=ConfigFolder.folder,
        separator=ConfigSeparator.separator,
        remove_duplicates=ConfigOps.remove_duplicates,
        remove_non_folders=ConfigOps.remove_non_folders,
        remove_non_abs=ConfigOps.remove_non_abs,
    )
    print(new_path)



"""
main entry point to the program
"""


import pylogconf.core
from pytconf.config import register_main, config_arg_parse_and_launch

from pypathutil.endpoints.group_default import register_group_default


def register_all_groups():
    """
    registers all groups of operations with pytconf
    """
    register_group_default()


@register_main()
def main():
    """
    pypathutil will help you manipulate paths
    """
    pylogconf.core.setup()
    register_all_groups()
    config_arg_parse_and_launch()


if __name__ == '__main__':
    main()
